"""
This module requires extra installs: ``pip install 'opendp[polars]'``

The ``opendp.extras.polars`` module adds differential privacy to the
`Polars DataFrame library <https://docs.pola.rs>`_.

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.polars``.
"""

from __future__ import annotations
from dataclasses import asdict, dataclass, field, replace
import os
from typing import Any, Literal, Mapping, Optional, Sequence, Union, cast
from opendp._lib import lib_path, import_optional_dependency
from opendp.extras.mbi import ContingencyTable, make_contingency_table, AIM, Algorithm
from opendp.mod import (
    ChangeOneIdDistance,
    Domain,
    Measurement,
    FrameDistance,
    Metric,
    OpenDPException,
    SymmetricIdDistance,
    binary_search,
    binary_search_chain,
)
from opendp.domains import (
    series_domain,
    lazyframe_domain,
    option_domain,
    atom_domain,
    categorical_domain,
    datetime_domain,
    enum_domain,
    array_domain,
)
from opendp.measurements import make_private_lazyframe
from deprecated import deprecated
from opendp.transformations import make_stable_lazyframe
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from opendp.context import Query
    from opendp.extras.polars.contingency_table import ContingencyTableQuery


class DPExpr(object):
    """
    If both ``opendp`` and ``polars`` have been imported,
    the methods of :py:class:`DPExpr` are registered under the ``dp`` namespace in
    `Polars expressions <https://docs.pola.rs/py-polars/html/reference/expressions/index.html>`_.
    An expression can be used as a plan in :py:func:`opendp.measurements.make_private_lazyframe`;
    See the full example there for more information.

    In addition to the DP-specific methods here, many Polars ``Expr`` methods are also supported,
    and are documented in the :ref:`API User Guide <expression-index>`.

    This class is typically not used directly by users:
    Instead its methods are registered under the ``dp`` namespace of Polars expressions.

    >>> import polars as pl
    >>> pl.len().dp
    <opendp.extras.polars.DPExpr object at ...>
    """

    def __init__(self, expr):
        self.expr = expr

    def noise(
        self,
        scale: float | None = None,
        distribution: Literal["Laplace"] | Literal["Gaussian"] | None = None,
    ):
        """Add noise to the expression.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.
        If distribution is None, then the noise distribution will be chosen for you:

        * Pure-DP: Laplace noise, where ``scale == standard_deviation / sqrt(2)``
        * zCDP: Gaussian noise, where ``scale == standard_devation``

        :param scale: Scale parameter for the noise distribution.
        :param distribution: Either Laplace, Gaussian or None.

        :example:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"A": list(range(100))}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(dp.len())
        >>> query.release().collect()
        shape: (1, 1)
        ┌─────┐
        │ len │
        │ --- │
        │ u32 │
        ╞═════╡
        │ ... │
        └─────┘
        """
        from polars.plugins import register_plugin_function  # type: ignore[import-not-found]
        from polars import lit  # type: ignore[import-not-found]

        return register_plugin_function(
            plugin_path=os.environ.get("OPENDP_POLARS_LIB_PATH", lib_path),
            function_name="noise",
            args=(self.expr, lit(distribution), scale),
            is_elementwise=True,
        )

    def laplace(self, scale: float | None = None):
        """Add Laplace noise to the expression.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: Noise scale parameter for the Laplace distribution. ``scale == standard_deviation / sqrt(2)``

        :example:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"A": list(range(100))}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(pl.len().dp.laplace())
        >>> query.release().collect()
        shape: (1, 1)
        ┌─────┐
        │ len │
        │ --- │
        │ u32 │
        ╞═════╡
        │ ... │
        └─────┘
        """
        return self.noise(scale=scale, distribution="Laplace")

    def gaussian(self, scale: float | None = None):
        """Add Gaussian noise to the expression.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: Noise scale parameter for the Gaussian distribution. ``scale == standard_deviation``

        :example:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"A": list(range(100))}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(rho=0.5),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(pl.len().dp.gaussian())
        >>> query.release().collect()
        shape: (1, 1)
        ┌─────┐
        │ len │
        │ --- │
        │ u32 │
        ╞═════╡
        │ ... │
        └─────┘
        """
        return self.noise(scale=scale, distribution="Gaussian")

    def len(self, scale: float | None = None):
        """Compute a differentially private estimate of the number of elements in `self`, including null values.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: parameter for the noise distribution.

        :example:

        This function is a shortcut for the exact Polars ``len`` and then noise addition:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"visits": [1, 2, None]}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(pl.col("visits").dp.len())
        >>> query.release().collect()
        shape: (1, 1)
        ┌────────┐
        │ visits │
        │ ---    │
        │ u32    │
        ╞════════╡
        │ ...    │
        └────────┘

        Output is noise added to three.

        It can differ from frame length (``.select(dp.len())``) if the expression uses transformations that change the number of rows,
        like filtering.
        """
        return self.expr.len().dp.noise(scale)

    def count(self, scale: float | None = None):
        """Compute a differentially private estimate of the number of elements in `self`, not including null values.

        This function is a shortcut for the exact Polars ``count`` and then noise addition.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: parameter for the noise distribution.

        :example:

        Count the number of records with known (non-null) visits:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"visits": [1, 2, None]}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(pl.col("visits").dp.count())
        >>> query.release().collect()
        shape: (1, 1)
        ┌────────┐
        │ visits │
        │ ---    │
        │ u32    │
        ╞════════╡
        │ ...    │
        └────────┘

        Output is noise added to three.
        """
        return self.expr.count().dp.noise(scale)

    def null_count(self, scale: float | None = None):
        """Compute a differentially private estimate of the number of null elements in `self`.

        This function is a shortcut for the exact Polars ``null_count`` and then noise addition.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: parameter for the noise distribution.

        :example:

        Count the number of records with unknown (null) visits:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"visits": [1, 2, None]}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(pl.col("visits").dp.null_count())
        >>> query.release().collect()
        shape: (1, 1)
        ┌────────┐
        │ visits │
        │ ---    │
        │ u32    │
        ╞════════╡
        │ ...    │
        └────────┘

        Output is noise added to one.

        Note that if you want to count the number of null *and* non-null records,
        consider combining the queries by constructing a boolean nullity column to group on,
        grouping by this column, and then using ``dp.len()``.
        """
        return self.expr.null_count().dp.noise(scale)

    def n_unique(self, scale: float | None = None):
        """Compute a differentially private estimate of the number of unique elements in `self`.

        This function is a shortcut for the exact Polars ``n_unique`` and then noise addition.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: parameter for the noise distribution.

        :example:

        Count the number of unique addresses:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"visits": [1, 2, None]}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ... )
        >>> query = context.query().select(pl.col("visits").dp.n_unique())
        >>> query.release().collect()
        shape: (1, 1)
        ┌────────┐
        │ visits │
        │ ---    │
        │ u32    │
        ╞════════╡
        │ ...    │
        └────────┘

        Output is noise added to three.
        """
        return self.expr.n_unique().dp.noise(scale)

    def sum(self, bounds: tuple[float, float], scale: float | None = None):
        """Compute the differentially private sum.

        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param bounds: clip the input data to these lower and upper bounds
        :param scale: parameter for the noise distribution

        :example:

        This function is a shortcut which actually implies several operations:

        * Clipping the values
        * Summing them
        * Applying noise to the sum

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"visits": [1, 2, None]}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ...     margins=[dp.polars.Margin(max_length=5)]
        ... )
        >>> query = context.query().select(pl.col("visits").fill_null(0).dp.sum((0, 1)))
        >>> query.release().collect()
        shape: (1, 1)
        ┌────────┐
        │ visits │
        │ ---    │
        │ i64    │
        ╞════════╡
        │ ...    │
        └────────┘

        Output is noise added to two due to each value being clipped to (0, 1).
        """
        return self.expr.clip(*bounds).sum().dp.noise(scale)

    def mean(
        self,
        bounds: tuple[float, float],
        scale: tuple[float | None, float | None] = (None, None),
    ):
        """Compute the differentially private mean.

        The amount of noise to be added to the sum is determined by the scale.
        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param bounds: clip the input data to these lower and upper bounds
        :param scale: parameters for the noise distributions of the numerator and denominator

        :example:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"visits": [1, 2, None]}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ...     margins=[dp.polars.Margin(max_length=5)]
        ... )
        >>> query = context.query().select(pl.col("visits").fill_null(0).dp.mean((0, 1)))
        >>> with pl.Config(float_precision=0): # just to prevent doctest from failing
        ...     query.release().collect()
        shape: (1, 1)
        ┌────────┐
        │ visits │
        │ ---    │
        │ f64    │
        ╞════════╡
        │ ...... │
        └────────┘

        Privately estimates the numerator and denominator separately, and then returns their ratio.
        """
        numer, denom = scale
        return self.sum(bounds, numer) / self.len(denom)

    def _discrete_quantile_score(self, alpha: float, candidates: list[float]):
        """Score the utility of each candidate for representing the true quantile.

        Candidates closer to the true quantile are assigned scores closer to zero.
        Lower scores are better.

        :param alpha: a value in [0, 1]. Choose 0.5 for median
        :param candidates: Set of possible quantiles to evaluate the utility of.
        """
        from polars.plugins import register_plugin_function  # type: ignore[import-not-found]
        from polars import Series  # type: ignore[import-not-found]

        return register_plugin_function(
            plugin_path=os.environ.get("OPENDP_POLARS_LIB_PATH", lib_path),
            function_name="discrete_quantile_score",
            args=[self.expr, alpha, Series(candidates)],
            returns_scalar=True,
            changes_length=True,
        )

    def _noisy_max(
        self,
        scale: float | None = None,
        negate: bool = False,
    ):
        """Report the noisy argmax or argmin.

        The scale calibrates the level of entropy when selecting an index.
        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param scale: Noise scale parameter for the Gumbel or Exponential distribution.
        :param negate: Enable to report noisy min.
        """
        from polars.plugins import register_plugin_function  # type: ignore[import-not-found]
        from polars import lit  # type: ignore[import-not-found]

        return register_plugin_function(
            plugin_path=os.environ.get("OPENDP_POLARS_LIB_PATH", lib_path),
            function_name="noisy_max",
            args=[self.expr, lit(negate), scale],
            is_elementwise=True,
        )

    def _index_candidates(self, candidates: list[float]):
        """Index into a candidate set.

        Typically used after :py:func:`_noisy_max` to map selected indices to candidates.

        :param candidates: The values that each selected index corresponds to.
        """
        from polars.plugins import register_plugin_function  # type: ignore[import-not-found]
        from polars import Series  # type: ignore[import-not-found]

        return register_plugin_function(
            plugin_path=os.environ.get("OPENDP_POLARS_LIB_PATH", lib_path),
            function_name="index_candidates",
            args=[self.expr, Series(candidates)],
            is_elementwise=True,
        )

    def quantile(
        self, alpha: float, candidates: list[float], scale: float | None = None
    ):
        """Compute a differentially private quantile.

        The scale calibrates the level of entropy when selecting a candidate.

        :param alpha: a value in [0, 1]. Choose 0.5 for median.
        :param candidates: Potential quantiles to select from.
        :param scale: How much noise to add to the scores of candidate.

        :example:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"age": list(range(100))}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ...     margins=[dp.polars.Margin(max_length=100)]
        ... )
        >>> candidates = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]
        >>> query = context.query().select(pl.col("age").cast(int).fill_null(0).dp.quantile(0.25, candidates))
        >>> query.release().collect()
        shape: (1, 1)
        ┌─────┐
        │ age │
        │ --- │
        │ i64 │
        ╞═════╡
        │ ... │
        └─────┘

        Output will be one of the candidates,
        with greater likelihood of being selected the closer the candidate is to the first quartile.
        """
        dq_score = self.expr.dp._discrete_quantile_score(alpha, candidates)
        noisy_idx = dq_score.dp._noisy_max(scale=scale, negate=True)
        return noisy_idx.dp._index_candidates(candidates)

    def median(self, candidates: list[float], scale: float | None = None):
        """Compute a differentially private median.

        The scale calibrates the level of entropy when selecting a candidate.
        If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

        :param candidates: Potential quantiles to select from.
        :param scale: How much noise to add to the scores of candidate.

        :example:

        >>> import polars as pl
        >>> context = dp.Context.compositor(
        ...     data=pl.LazyFrame({"age": list(range(100))}),
        ...     privacy_unit=dp.unit_of(contributions=1),
        ...     privacy_loss=dp.loss_of(epsilon=1.),
        ...     split_evenly_over=1,
        ...     margins=[dp.polars.Margin(max_length=100)]
        ... )
        >>> candidates = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]
        >>> query = context.query().select(pl.col("age").cast(int).fill_null(0).dp.quantile(0.5, candidates))
        >>> query.release().collect()
        shape: (1, 1)
        ┌─────┐
        │ age │
        │ --- │
        │ i64 │
        ╞═════╡
        │ ... │
        └─────┘

        Output will be one of the candidates,
        with greater likelihood of being selected the closer the candidate is to the median.
        """
        return self.expr.dp.quantile(0.5, candidates, scale)


pl = import_optional_dependency("polars", raise_error=False)
if pl is not None:
    pl.api.register_expr_namespace("dp")(DPExpr)


def dp_len(scale: float | None = None):
    """Compute a differentially private estimate of the number of rows.

    If scale is None it is filled by ``global_scale`` in :py:func:`opendp.measurements.make_private_lazyframe`.

    :param scale: parameter for the noise distribution.

    :example:

    This function is a shortcut for the exact Polars ``len`` and then noise addition:

    >>> import polars as pl
    >>> context = dp.Context.compositor(
    ...     data=pl.LazyFrame({"A": list(range(100))}),
    ...     privacy_unit=dp.unit_of(contributions=1),
    ...     privacy_loss=dp.loss_of(epsilon=1.),
    ...     split_evenly_over=1,
    ... )
    >>> query = context.query().select(dp.len())
    >>> query.release().collect()
    shape: (1, 1)
    ┌─────┐
    │ len │
    │ --- │
    │ u32 │
    ╞═════╡
    │ ... │
    └─────┘
    """
    from polars.functions import len  # type: ignore[import-not-found]

    return DPExpr(len()).noise(scale=scale)


class OnceFrame(object):
    """OnceFrame is a Polars LazyFrame that may only be collected into a DataFrame once.

    The APIs on this class mimic those that can be found in Polars.

    Differentially private guarantees on a given LazyFrame require the LazyFrame to be evaluated at most once.
    The purpose of this class is to protect against repeatedly evaluating the LazyFrame.
    """

    def __init__(self, queryable):
        self.queryable = queryable

    def collect(self):
        """Collects a DataFrame from a OnceFrame, exhausting the OnceFrame."""
        from opendp._data import onceframe_collect

        return onceframe_collect(self.queryable)

    def lazy(self):
        """Extracts a ``LazyFrame`` from a ``OnceFrame``,
        circumventing protections against multiple evaluations.

        Each collection consumes the entire allocated privacy budget.
        To remain DP at the advertised privacy level, only collect the ``LazyFrame`` once.

        Requires "honest-but-curious" because the privacy guarantees only apply if:

        1. The LazyFrame (compute plan) is only ever executed once.
        2. The analyst does not observe ordering of rows in the output.

        To ensure that row ordering is not observed:

        1. Do not extend the compute plan with order-sensitive computations.
        2. Shuffle the output once collected `(in Polars sample all, with shuffle enabled) <https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.sample.html>`_.
        """
        from opendp._data import onceframe_lazy

        return onceframe_lazy(self.queryable)


def _lazyframe_domain_from_schema(schema) -> Domain:
    """Builds the broadest possible LazyFrameDomain that matches a given LazyFrame schema."""
    return lazyframe_domain(
        [_series_domain_from_field(field) for field in schema.items()]
    )


def _series_domain_from_field(field) -> Domain:
    """Builds the broadest possible SeriesDomain that matches a given field."""
    name, dtype = field
    return series_domain(name, option_domain(_domain_from_dtype(dtype)))


def _domain_from_dtype(dtype) -> Domain:
    """Builds the broadest possible Domain that matches a given dtype."""
    import polars as pl

    if dtype == pl.Categorical:
        return categorical_domain()
    if dtype == pl.Enum:
        return enum_domain(dtype.categories)
    if dtype == pl.Datetime:
        return datetime_domain(dtype.time_unit, dtype.time_zone)
    if dtype == pl.Array:
        return array_domain(_domain_from_dtype(dtype.inner), dtype.size)

    T = {
        pl.UInt32: "u32",
        pl.UInt64: "u64",
        pl.Int8: "i8",
        pl.Int16: "i16",
        pl.Int32: "i32",
        pl.Int64: "i64",
        pl.Float32: "f32",
        pl.Float64: "f64",
        pl.Boolean: "bool",
        pl.String: "String",
        pl.Time: "NaiveTime",
        pl.Date: "NaiveDate",
    }.get(dtype)

    if T is None:
        raise ValueError(f"unrecognized dtype: {dtype}")  # pragma: no cover

    return atom_domain(T=T)


_LAZY_EXECUTION_METHODS = {
    "collect",
    "collect_async",
    "describe",
    "sink_parquet",
    "sink_ipc",
    "sink_csv",
    "sink_ndjson",
    "fetch",
}


@dataclass
class SortBy:
    """Configuration for ``keep`` in :py:meth:`LazyFrameQuery.truncate_per_group`.

    Follows the arguments in Polars' ``sort_by`` method.
    """

    by: Any
    """Column(s) to sort by. Accepts expression input. Strings are parsed as column names."""

    descending: bool | Sequence[bool] = False
    """Sort in descending order. When sorting by multiple columns, can be specified per column by passing a sequence of booleans."""

    nulls_last: bool | Sequence[bool] = False
    """Place null values last; can specify a single boolean applying to all columns or a sequence of booleans for per-column control."""

    multithreaded: bool = True
    """Sort using multiple threads."""

    maintain_order: bool = False
    """Whether the order should be maintained if elements are equal."""


class LazyFrameQuery:
    """
    A ``LazyFrameQuery`` may be returned by :py:func:`opendp.context.Context.query`.
    It mimics a `Polars LazyFrame <https://docs.pola.rs/api/python/stable/reference/lazyframe/index.html>`_,
    but makes a few additions and changes as documented below."""

    # Keep this docstring in sync with the docstring below for the dummy class.

    def __init__(self, polars_plan, query):
        self.polars_plan = polars_plan
        self._query = query
        # do not initialize super() because inheritance is only used to mimic the API surface

    def __getattribute__(self, name):
        # Re-route all possible attribute access to self.polars_plan.
        # __getattribute__ is necessary because __getattr__ cannot intercept calls to inherited methods

        # We keep the query plan void of data anyways,
        # so running the computation doesn't affect privacy.
        # This doesn't have to cover all possible APIs that may execute the query,
        #    but it does give a simple sanity check for the obvious cases.
        if name in _LAZY_EXECUTION_METHODS:
            raise ValueError("You must call `.release()` before executing a query.")

        polars_plan = object.__getattribute__(self, "polars_plan")
        query = object.__getattribute__(self, "_query")
        attr = getattr(polars_plan, name, None)

        # If not a valid attribute on self.polars_plan, then don't re-route
        if attr is None:
            return object.__getattribute__(self, name)

        # any callable attributes (like .with_columns or .select) will now also wrap their outputs in a LazyFrameQuery
        if callable(attr):

            def _wrap(*args, **kwargs):
                out = attr(*args, **kwargs)

                if pl is not None:
                    # re-wrap any lazy outputs to keep the conveniences afforded by this class
                    if isinstance(out, pl.lazyframe.frame.LazyFrame):
                        return LazyFrameQuery(out, query)

                    if isinstance(out, pl.lazyframe.group_by.LazyGroupBy):
                        return LazyGroupByQuery(out, query)

                return out

            return _wrap
        return attr

    # These definitions are primarily for mypy:
    # Without them, a "# type: ignore[union-attr]" is needed on every line where these methods are used.
    # The docstrings are not seen by Sphinx, but aren't doing any harm either.

    def sort(  # type: ignore[empty-body]
        self,
        by,
        *more_by,
        descending: bool | Sequence[bool] = False,
        nulls_last: bool | Sequence[bool] = False,
        maintain_order: bool = False,
        multithreaded: bool = True,
    ) -> LazyFrameQuery:
        """Sort the ``LazyFrame`` by the given columns."""
        ...

    def filter(  # type: ignore[empty-body]
        self,
        *predicates,
        **constraints: Any,
    ) -> LazyFrameQuery:
        """
        Filter the rows in the ``LazyFrame`` based on a predicate expression.

        OpenDP discards relevant margin descriptors in the domain when filtering.
        """
        ...

    def select(  # type: ignore[empty-body]
        self, *exprs, **named_exprs
    ) -> LazyFrameQuery:
        """
        Select columns from this ``LazyFrame``.

        OpenDP expects expressions in select statements that don't aggregate to be row-by-row.
        """
        ...

    def select_seq(  # type: ignore[empty-body]
        self, *exprs, **named_exprs
    ) -> LazyFrameQuery:
        """
        Select columns from this ``LazyFrame``.

        OpenDP allows expressions in select statements that aggregate to not be row-by-row.
        """
        ...

    def group_by(  # type: ignore[empty-body]
        self,
        *by,
        maintain_order: bool = False,
        **named_by,
    ) -> LazyGroupByQuery:
        """
        Start a group by operation.

        OpenDP currently requires that grouping keys be simple column expressions.
        """
        ...

    def with_columns(  # type: ignore[empty-body]
        self,
        *exprs,
        **named_exprs,
    ) -> LazyFrameQuery:
        """
        Add columns to this ``LazyFrame``.

        OpenDP requires that expressions in with_columns are row-by-row:
        expressions may not change the number or order of records
        """
        ...

    def with_columns_seq(  # type: ignore[empty-body]
        self,
        *exprs,
        **named_exprs,
    ) -> LazyFrameQuery:
        """
        Add columns to this ``LazyFrame``.

        OpenDP requires that expressions in with_columns are row-by-row:
        expressions may not change the number or order of records
        """
        ...

    def join(  # type: ignore[empty-body]
        self,
        other,
        on=None,
        how="inner",
        *,
        left_on=None,
        right_on=None,
        suffix: str = "_right",
        validate="m:m",
        join_nulls: bool = False,
        coalesce: bool | None = None,
        allow_parallel: bool = True,
        force_parallel: bool = False,
    ) -> LazyFrameQuery:
        """
        Add a join operation to the Logical Plan.
        """
        ...

    def with_keys(
        self,
        keys,
        on: list[str] | None = None,
    ) -> LazyFrameQuery:
        """
        Shorthand to join with an explicit key-set.

        :param keys: lazyframe containing a key-set whose columns correspond to the grouping keys
        :param on: optional, the names of columns to join on. Useful if the key dataframe contains extra columns
        """
        # Motivation for adding this new API:
        # 1. Writing a left join is more difficult in the context API:
        #   see the complexity of this implementation, where you have to go under the hood.
        #   This gives an easier shorthand to write a left join.
        # 2. Left joins are more likely to be supported by database backends.
        # 3. Easier to use; with the Polars API the key set needs to be lazy, user must specify they want a right join and the join keys.

        if pl is not None:
            if isinstance(keys, pl.dataframe.frame.DataFrame):
                keys = keys.lazy()

        if on is None:
            on = keys.collect_schema().names()

        return LazyFrameQuery(
            keys.join(self.polars_plan, how="left", on=on, nulls_equal=True),
            self._query,
        )

    def truncate_per_group(
        self,
        k: int,
        by: list[Any] | None = None,
        keep: Literal["sample", "first", "last"] | SortBy = "sample",
    ) -> LazyFrameQuery:
        """
        Limit the number of contributed rows per group.

        :param k: the number of rows to keep for each identifier and group
        :param by: optional, additional columns to group by
        :param keep: which rows to keep for each identifier in each group
        """
        input_metric = self._query._chain[1]

        if isinstance(by, str):
            raise ValueError(
                "by must be a list of strings or expressions"
            )  # pragma: no cover

        if isinstance(input_metric, FrameDistance):
            input_metric = input_metric.inner_metric
        if not isinstance(input_metric, (SymmetricIdDistance, ChangeOneIdDistance)):
            raise ValueError("truncation is only valid when identifier is defined")

        if keep == "sample":
            indexes = pl.int_range(pl.len()).shuffle()
        elif keep == "first":
            indexes = pl.int_range(pl.len())
        elif keep == "last":
            indexes = pl.int_range(pl.len()).reverse()
        elif isinstance(keep, SortBy):
            indexes = pl.int_range(pl.len()).sort_by(**asdict(keep))
        else:
            raise ValueError(
                "keep must be 'sample', 'first', 'last' or SortBy"
            )  # pragma: no cover

        return self.filter(indexes.over(input_metric.identifier, *by or []) < k)

    def truncate_num_groups(
        self,
        k: int,
        by: list[Any],
        keep: Literal["sample", "first", "last"] = "sample",
    ) -> LazyFrameQuery:
        """
        Limit the number of groups an individual may influence.

        :param k: the number of groups to keep for each identifier
        :param by: when grouped by these grouping columns
        :param keep: which groups to keep for each identifier
        """
        input_metric = self._query._chain[1]

        if isinstance(by, str):
            raise ValueError(
                "by must be a list of strings or expressions"
            )  # pragma: no cover

        if isinstance(input_metric, FrameDistance):
            input_metric = input_metric.inner_metric
        if not isinstance(input_metric, (SymmetricIdDistance, ChangeOneIdDistance)):
            raise ValueError("truncation is only valid when identifier is defined")

        struct = pl.struct(*by)
        if keep == "sample":
            ranks = pl.struct(struct.hash(), struct).rank("dense")
        elif keep == "first":
            ranks = struct.rank("dense", descending=False)
        elif keep == "last":
            ranks = struct.rank("dense", descending=True)
        else:
            raise ValueError(
                "keep must be 'sample', 'first' or 'last'"
            )  # pragma: no cover

        return self.filter(ranks.over(input_metric.identifier) < k)

    def resolve(self) -> Measurement:
        """Resolve the query into a measurement."""

        # access attributes of self without getting intercepted by Self.__getattribute__
        polars_plan = object.__getattribute__(self, "polars_plan")
        query = object.__getattribute__(self, "_query")
        input_domain, input_metric = query._chain
        d_in, d_out = query._d_in, query._d_out

        def _make(scale, threshold=None):
            return make_private_lazyframe(
                input_domain=input_domain,
                input_metric=input_metric,
                output_measure=query._output_measure,
                lazyframe=polars_plan,
                global_scale=scale,
                threshold=threshold,
            )

        # when the query has sensitivity zero or is behind an invariant
        try:
            m_zero = _make(0.0, threshold=None)
            if m_zero.check(d_in, d_out):
                # if the zero scale measurement is already private, return it
                return m_zero
        except OpenDPException:
            pass

        # when the output measure is δ-approximate, then there are two free parameters to tune
        if getattr(query._output_measure.type, "origin", None) == "Approximate":
            # search for a scale parameter. Solve for epsilon first,
            # setting threshold to u32::MAX so as not to interfere with the search for a suitable scale parameter
            scale = binary_search(
                lambda s: _make(s, threshold=2**32 - 1).map(d_in)[0] < d_out[0],  # type: ignore[index]
                T=float,
            )

            # attempt to return without setting a threshold
            try:
                return _make(scale, threshold=None)
            except OpenDPException:
                pass

            # now that scale has been solved, find a suitable threshold
            threshold = binary_search(
                lambda t: _make(scale, t).map(d_in)[1] < d_out[1],  # type: ignore[index]
                T=int,
            )

            # return a measurement with the discovered scale and threshold
            return _make(scale, threshold)

        # when no delta parameter is involved,
        # finding a suitable measurement just comes down to finding scale
        return binary_search_chain(_make, d_in, d_out, T=float)

    def release(self) -> OnceFrame:
        """Release the query. The query must be part of a context."""
        query = object.__getattribute__(self, "_query")
        resolve = object.__getattribute__(self, "resolve")
        return query._context(resolve())  # type: ignore[misc]

    def summarize(self, alpha: float | None = None):
        """Summarize the statistics released by this query.

        If ``alpha`` is passed, the resulting data frame includes an ``accuracy`` column.

        If a threshold is configured for censoring small/sensitive groups,
        a threshold column will be included,
        containing the cutoff for the respective count query being thresholded.

        :param alpha: optional. A value in [0, 1] denoting the statistical significance. For the corresponding confidence level, subtract from from 1: for 95% confidence, use 0.05 for alpha.

        :example:

        .. code:: pycon

            >>> import polars as pl
            >>> data = pl.LazyFrame([pl.Series("convicted", [0, 1, 1, 0, 1] * 50, dtype=pl.Int32)])

            >>> context = dp.Context.compositor(
            ...     data=data,
            ...     privacy_unit=dp.unit_of(contributions=1),
            ...     privacy_loss=dp.loss_of(epsilon=1.0),
            ...     split_evenly_over=1,
            ...     margins=[dp.polars.Margin(by=(), max_length=1000)],
            ... )

            >>> query = context.query().select(
            ...     dp.len(),
            ...     pl.col("convicted").fill_null(0).dp.sum((0, 1))
            ... )

            >>> query.summarize(alpha=.05)  # type: ignore[union-attr]
            shape: (2, 5)
            ┌───────────┬──────────────┬─────────────────┬───────┬──────────┐
            │ column    ┆ aggregate    ┆ distribution    ┆ scale ┆ accuracy │
            │ ---       ┆ ---          ┆ ---             ┆ ---   ┆ ---      │
            │ str       ┆ str          ┆ str             ┆ f64   ┆ f64      │
            ╞═══════════╪══════════════╪═════════════════╪═══════╪══════════╡
            │ len       ┆ Frame Length ┆ Integer Laplace ┆ 2.0   ┆ 6.429605 │
            │ convicted ┆ Sum          ┆ Integer Laplace ┆ 2.0   ┆ 6.429605 │
            └───────────┴──────────────┴─────────────────┴───────┴──────────┘

        The accuracy in any given row can be interpreted with:

        .. code:: pycon

            >>> def interpret_accuracy(distribution, scale, accuracy, alpha):
            ...     return (
            ...         f"When the {distribution} scale is {scale}, "
            ...         f"the DP estimate differs from the true value by no more than {accuracy} "
            ...         f"at a statistical significance level alpha of {alpha}, "
            ...         f"or with (1 - {alpha})100% = {(1 - alpha) * 100}% confidence."
            ...     )

            >>> interpret_accuracy("Integer Laplace", 2.0, 6.429605, alpha=.05) # doctest:+SKIP
        """
        from opendp.accuracy import summarize_polars_measurement

        return summarize_polars_measurement(self.resolve(), alpha)

        
    def contingency_table(
        self,
        *,
        keys: Optional[Mapping[str, Sequence]] = None,
        cuts: Optional[Mapping[str, Sequence[float]]] = None,
        table: Optional[ContingencyTable] = None,
        algorithm: Union[Algorithm] = AIM(),
    ) -> "ContingencyTableQuery":
        """Release an approximation to a contingency table across all columns.

        :param keys: dictionary of column names and unique categories
        :param cuts: dictionary of column names and bin edges for numerical columns
        :param table: ContingencyTable from prior release
        :param algorithm: configuration for internal estimation algorithm
        """
        from .contingency_table import ContingencyTableQuery

        query: Query = object.__getattribute__(self, "_query")
        input_domain, input_metric = cast(tuple[Domain, Metric], query._chain)
        d_in, d_out = query._d_in, query._d_out

        t_plan = make_stable_lazyframe(
            input_domain,
            input_metric,
            lazyframe=object.__getattribute__(self, "polars_plan"),
        )

        m_table, oneway_scale, oneway_threshold = make_contingency_table(
            input_domain=t_plan.output_domain,
            input_metric=t_plan.output_metric,
            output_measure=query._output_measure,
            d_in=t_plan.map(d_in),
            d_out=d_out,  # type: ignore[arg-type]
            keys=keys,
            cuts=cuts,
            table=table,
            algorithm=algorithm,
        )

        return ContingencyTableQuery(
            chain=t_plan >> m_table,
            output_measure=query._output_measure,
            context=query._context,
            oneway_scale=oneway_scale,
            oneway_threshold=oneway_threshold
        )


class LazyGroupByQuery:
    """
    A ``LazyGroupByQuery`` is returned by :py:func:`opendp.extras.polars.LazyFrameQuery.group_by`.
    It mimics a `Polars LazyGroupBy <https://docs.pola.rs/api/python/stable/reference/lazyframe/group_by.html>`_,
    but only supports APIs documented below."""

    def __init__(self, lgb_plan, query):
        self._lgb_plan = lgb_plan
        self._query = query

    def agg(
        self,
        *aggs,
        **named_aggs,
    ) -> LazyFrameQuery:
        """
        Compute aggregations for each group of a group by operation.

        :param aggs: expressions to apply in the aggregation context
        :param named_aggs: named/aliased expressions to apply in the aggregation context
        """
        polars_plan = self._lgb_plan.agg(*aggs, **named_aggs)
        return LazyFrameQuery(polars_plan, self._query)


@dataclass
class Margin:
    """
    The ``Margin`` class is used to describe what information is known publicly about a grouped dataset:
    like the values you might expect to find in the margins of a table.

    Be aware that aspects of your data marked as "public information" are not subject to privacy protections,
    so it is important that public descriptors about the margin should be set conservatively, or not set at all.

    Instances of this class are used by :py:func:`opendp.context.Context.compositor`.
    """

    by: Sequence = field(default_factory=list)
    """Polars expressions describing the grouping columns."""

    max_length: int | None = None
    """An upper bound on the number of records in any one group.

    Some operations (for instance, for float sums) will error if `max_length` is not provided.
    This is used to resolve issues raised in the paper
    `Widespread Underestimation of Sensitivity in Differentially Private Libraries and How to Fix It <https://arxiv.org/pdf/2207.10635.pdf>`_.
    
    If you don't know how many records are in the data, you can specify a very loose upper bound,
    for example, the size of the total population you are sampling from.
    """

    max_groups: int | None = None
    """An upper bound on the number of distinct groups."""

    invariant: Literal["keys"] | Literal["lengths"] | None = None
    """Identifies properties of grouped data that are considered invariant.
    
    * ``"keys"`` designates that keys are not protected
    * ``"lengths"`` designates that both keys and group lengths are not protected

    By the analysis of invariants conducted in
    `Formal Privacy Guarantees with Invariant Statistics <https://arxiv.org/abs/2410.17468>`_,
    when invariants are set, the effective privacy guarantees of the library are weaker than advertised.
    """

    @property
    @deprecated(version="0.13.0", reason="Use max_length instead.")
    def max_partition_length(self):
        return self.max_length  # pragma: no cover

    @max_partition_length.setter
    @deprecated(reason="Use max_length instead.")
    def max_partition_length(self, value):
        self.max_length = value  # pragma: no cover

    @property
    @deprecated(
        version="0.13.0",
        reason="Use max_groups instead. This was renamed to be consistent with Polars terminology."
    )
    def max_num_partitions(self):
        return self.max_groups  # pragma: no cover

    @max_num_partitions.setter
    @deprecated(
        version="0.13.0",
        reason="Use max_groups instead. This was renamed to be consistent with Polars terminology."
    )
    def max_num_partitions(self, value):
        self.max_groups = value  # pragma: no cover

    @property
    @deprecated(
        version="0.13.0",
        reason='Use invariant instead. This was renamed because invariants are not "public information". Invariants are "unprotected information".'
    )
    def public_info(self):
        return self.invariant  # pragma: no cover

    @public_info.setter
    @deprecated(
        version="0.13.0",
        reason='Use invariant instead. This was renamed because invariants are not "public information". Invariants are "unprotected information".'
    )
    def public_info(self, value):
        self.invariant = value  # pragma: no cover

    @property
    def max_partition_contributions(self):
        raise NotImplementedError(
            "max_partition_contributions has been moved. Use `dp.unit_of(contributions=[dp.polars.Bound(per_group=...)])` instead."
        )  # pragma: no cover

    @max_partition_contributions.setter
    def max_partition_contributions(self, value):
        raise NotImplementedError(
            "max_partition_contributions has been moved. Use `dp.unit_of(contributions=[dp.polars.Bound(per_group=...)])` instead."
        )  # pragma: no cover

    @property
    def max_influenced_partitions(self):
        raise NotImplementedError(
            "max_influenced_partitions has been moved. Use `dp.unit_of(contributions=[dp.polars.Bound(num_groups=...)])` instead."
        )  # pragma: no cover

    @max_influenced_partitions.setter
    def max_influenced_partitions(self, value):
        raise NotImplementedError(
            "max_influenced_partitions has been moved. Use `dp.unit_of(contributions=[dp.polars.Bound(num_groups=...)])` instead."
        )  # pragma: no cover

    def __eq__(self, other) -> bool:
        if not isinstance(other, Margin):
            return False

        # special logic for by, which is considered a set (order and dupes don't matter)
        # and may contain expressions that do not have a boolean equality operator
        def serialize(by):
            if isinstance(by, str):
                by = pl.col(by)
            return by.meta.serialize()

        self_by = {serialize(col) for col in self.by}
        other_by = {serialize(col) for col in other.by}
        if self_by != other_by:
            return False

        return asdict(replace(self, by=[])) == asdict(replace(other, by=[]))


@dataclass
class Bound(object):
    """
    The ``Bound`` class is used to describe bounds on the number of
    contributed rows per-group and the number of contributed groups.
    """

    by: Sequence = field(default_factory=list)
    """Polars expressions describing the grouping columns."""

    per_group: int | None = None
    """The greatest number of records an individual may contribute to any one group.
    
    This can significantly reduce the sensitivity of grouped queries under zero-Concentrated DP.
    """

    num_groups: int | None = None
    """The greatest number of groups any one individual can contribute."""

    def __eq__(self, other) -> bool:
        if not isinstance(other, Bound):
            return False

        # special logic for by, which is considered a set (order and dupes don't matter)
        # and may contain expressions that do not have a boolean equality operator
        def serialize(by):
            if isinstance(by, str):
                by = pl.col(by)
            return by.meta.serialize()

        self_by = {serialize(col) for col in self.by}
        other_by = {serialize(col) for col in other.by}
        if self_by != other_by:
            return False

        return asdict(replace(self, by=[])) == asdict(replace(other, by=[]))  # type: ignore[arg-type]



from opendp._convert import *
from opendp._convert import (
    _scalar_to_slice, _slice_to_scalar,
    _vector_to_slice, _slice_to_vector,
    _hashmap_to_slice, _slice_to_hashmap,
)
from opendp._convert import _check_and_cast_scalar
from opendp.typing import *
import pytest


def test_data_object_int():
    val_in = 123
    obj = py_to_c(val_in, c_type=AnyObjectPtr)
    val_out = c_to_py(obj)
    assert val_out == val_in


def test_data_object_int_to_float():
    val_in = 123
    obj = py_to_c(val_in, c_type=AnyObjectPtr, type_name='f64')
    val_out = c_to_py(obj)
    assert val_out == val_in


def test_data_object_float():
    val_in = 123.123
    obj = py_to_c(val_in, c_type=AnyObjectPtr)
    val_out = c_to_py(obj)
    assert val_out == val_in


def test_data_object_str():
    val_in = "hello, world"
    obj = py_to_c(val_in, c_type=AnyObjectPtr)
    val_out = c_to_py(obj)
    assert val_out == val_in


def test_data_object_list():
    val_in = [1, 2, 3]
    obj = py_to_c(val_in, c_type=AnyObjectPtr)
    val_out = c_to_py(obj)
    assert val_out == val_in


def test_data_object_tuple():
    val_in = (1., 1e-7)
    obj = py_to_c(val_in, c_type=AnyObjectPtr)
    val_out = c_to_py(obj)
    assert val_out == val_in


def test_data_object_string_pointer():
    val_in = "hello, world"
    obj = py_to_c(val_in, c_type=ctypes.c_void_p, type_name='String')
    assert obj.value == val_in.encode()


def test_roundtrip_int():
    in_ = 23
    ptr = ctypes.POINTER(ctypes.c_int32)(ctypes.c_int32(in_))
    ptr = ctypes.byref(ctypes.c_int32(in_))  # type: ignore[assignment]

    int_void = ctypes.cast(ptr, ctypes.c_void_p)
    out = ctypes.cast(int_void, ctypes.POINTER(ctypes.c_int32)).contents.value
    assert in_ == out


def test_roundtrip_ffislice_int():
    in_ = 23
    ptr = ctypes.byref(ctypes.c_int32(in_))
    ffi_slice = FfiSlice(ctypes.cast(ptr, ctypes.c_void_p), 1)
    out = ctypes.cast(ffi_slice.ptr, ctypes.POINTER(ctypes.c_int32)).contents.value
    assert out == in_


def test_roundtrip_ffisliceptr_int():
    in_ = 23
    ptr = ctypes.byref(ctypes.c_int32(in_))
    ffi_slice = FfiSlice(ctypes.cast(ptr, ctypes.c_void_p), 1)
    ffi_slice_ptr = FfiSlicePtr(ffi_slice)
    ffi_slice = ffi_slice_ptr.contents
    out = ctypes.cast(ffi_slice.ptr, ctypes.POINTER(ctypes.c_int32)).contents.value
    assert out == in_


def test_roundtrip_ffisliceptr_int_lib():
    in_ = 23
    ffi_slice_ptr = _scalar_to_slice(in_, 'i32')
    out = _slice_to_scalar(ffi_slice_ptr, 'i32')
    assert out == in_


def test_vec_str():
    data = ["a", "bbb", "c"]
    # partial roundtrip
    slice = _vector_to_slice(data, type_name=Vec[str])
    assert _slice_to_vector(slice, type_name=Vec[str]) == data

    # complete roundtrip
    any = py_to_c(data, c_type=AnyObjectPtr)
    assert c_to_py(any) == data


def test_hashmap():
    data = {"A": 23, "B": 12, "C": 234}
    slice = _hashmap_to_slice(data, HashMap[str, int])
    assert _slice_to_hashmap(slice) == data

    # complete roundtrip
    any = py_to_c(data, c_type=AnyObjectPtr)
    assert c_to_py(any) == data


def test_numpy_data():
    np = pytest.importorskip('numpy')
    def roundtrip(value, type_name, dtype=None):
        assert value == c_to_py(py_to_c(np.array(value, dtype=dtype), AnyObjectPtr, type_name=type_name))
    roundtrip([1, 2], "Vec<i32>", dtype=np.int32)
    roundtrip(1, "i32", dtype=np.int32)
    roundtrip([1., 2.], "Vec<f64>")
    roundtrip(1., "f64")
    roundtrip(["A", "B"], "Vec<String>")
    roundtrip("A", "String")


def test_numpy_trans():
    np = pytest.importorskip('numpy')
    import opendp.prelude as dp
    assert dp.t.make_sum(
        dp.vector_domain(dp.atom_domain(bounds=(0, 10))), 
        dp.symmetric_distance(),
    )(np.array([1, 2, 3], dtype=np.int32)) == 6


def test_overflow():
    import pytest
    with pytest.raises(ValueError):
        py_to_c(-1, AnyObjectPtr, u8)

    with pytest.raises(ValueError):
        py_to_c(256, AnyObjectPtr, u8)
  
    with pytest.raises(ValueError):
        py_to_c(-129, AnyObjectPtr, i8)


def test_polars_dataframe():
    pl = pytest.importorskip("polars")
    val_in = pl.DataFrame({
        "A": [1] * 100,
        "B": ["X"] * 100,
        "C": [True] * 100,
    })
    obj = py_to_c(val_in, AnyObjectPtr, "DataFrame")
    val_out = c_to_py(obj)
    assert val_out.equals(val_in)


def test_polars_expr():
    pl = pytest.importorskip("polars")
    val_in = pl.all()
    obj = py_to_c(val_in, AnyObjectPtr, "Expr")
    val_out = c_to_py(obj)
    assert str(val_out) == str(val_in)


def test_check_and_cast_scalar():
    # Any number is cast to float, if f32 or f64 is expected:
    assert _check_and_cast_scalar('f32', 1.0) == 1.0
    assert _check_and_cast_scalar('f64', 1) == 1.0

    # Ints cast to ints, unsurprisingly:
    assert _check_and_cast_scalar('u8', 1) == 1

    # There are range checks for ints:
    with pytest.raises(ValueError, match="256 is not representable by u8"):
        _check_and_cast_scalar('u8', 256)

    # Floats don't have range checks because inf is a valid float:
    assert _check_and_cast_scalar('f32', 1e100) == 1e+100

    # Floats cannot be cast to ints:
    with pytest.raises(TypeError, match="inferred type is f64, expected i32."):
        _check_and_cast_scalar('i32', 1.0)
    
    # Bools can only cast to bools:
    assert _check_and_cast_scalar('bool', True)

    # Bools cannot cast to ints:
    with pytest.raises(TypeError, match="inferred type is bool, expected u8."):
        _check_and_cast_scalar('u8', True)
    
    # Bools cannot cast to floats:
    with pytest.raises(TypeError, match="inferred type is bool, expected f64."):
        _check_and_cast_scalar('f64', True)
    
    with pytest.raises(TypeError, match="inferred type is i32, expected bool."):
        _check_and_cast_scalar('bool', 1)
    
    with pytest.raises(TypeError, match="inferred type is i32, expected fake."):
        _check_and_cast_scalar('fake', 1)

def test_bitvec():
    np = pytest.importorskip('numpy') 

    for i in range(1, 20):
        data = np.packbits([1] * i)
        obj = py_to_c(data.tobytes(), AnyObjectPtr, "BitVector")
        val_out = np.frombuffer(c_to_py(obj), dtype=np.uint8)
        bits = np.unpackbits(val_out)
        assert (bits.tolist() + [0]).index(0) == i




# type: ignore
def sample_geometric_buffer(
    buffer_len: usize, constant_time: bool
) -> Optional[uint]:  # |\label{line:geombuffer}|
    if constant_time:
        buf = bytearray(buffer_len)
        fill_bytes(buf)  # mutates in-place
        ret = None
        for i in range(buffer_len):
            # find first nonzero event
            if buf[i] > 0:
                # compute index of first nonzero bit buffer
                cand = 8 * i + buf[i].leading_zeroes()  # |\label{line:indexcmp}|
                ret = cand if ret is None else min(ret, cand)
        return ret
    else:
        for i in range(buffer_len):
            buf = bytearray(1)
            fill_bytes(buf)  # mutates in-place
            if buf[0] > 0:
                return 8 * i + buf[0].leading_zeroes()

        return None



'''
This module requires extra installs: ``pip install 'opendp[scikit-learn]'``

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.sklearn``.
Submodule organization will follow the conventions of `scikit-learn <https://scikit-learn.org/stable/api/index.html>`_.
'''

import opendp.extras.sklearn.decomposition as decomposition
import opendp.extras.sklearn.linear_model as linear_model


# type: ignore
def make_int_to_bigint(
    input_space: tuple[VectorDomain[AtomDomain[T]], LpDistance[P, QI]],
) -> Transformation[
    VectorDomain[AtomDomain[T]],
    VectorDomain[AtomDomain[IBig]],
    LpDistance[P, QI],
    LpDistance[P, RBig],
]:
    input_domain, input_metric = input_space

    def stability_map(d_in):
        try:
            return RBig.try_from(d_in)
        except Exception:
            raise f"d_in ({d_in}) must be finite"

    return Transformation.new(
        input_domain,
        VectorDomain(
            element_domain=AtomDomain.default(IBig),
            size=input_domain.size,
        ),
        Function.new(lambda x: [IBig.from_(x_i) for x_i in x]),
        input_metric,
        LpDistance.default(),
        StabilityMap.new_fallible(stability_map),
    )



# type: ignore
def make_randomized_response_bitvec(
        input_domain: BitVectorDomain, 
        input_metric: DiscreteDistance, 
        f: f64, 
        constant_time: bool):
    output_measure = MaxDivergence(f64)
    
    if f <= 0.0 or f > 1: # |\label{line:range}|
        raise Exception("Probability must be in (0.0, 1]")
    
    m = input_domain.max_weight
    if m is None: 
        raise Exception("make_randomized_response_bitvec requires a max_weight on the input_domain")
    
    epsilon = 2 * m * log((2 - f) / f)
    def privacy_map(d_in: IntDistance): # |\label{line:map}|
        if d_in == 0:
            return 0.
        if d_in > 1:
            raise ValueError("d_in must be 0 or 1.")
        return epsilon
    def function(arg: BitVector) -> BitVector: # |\label{line:fn}|
        k = len(arg)
        noise_vector = [bool.sample_bernoulli(f/2, constant_time) for _ in range(k)]
        return xor(arg, noise_vector)
    
    return Measurement(input_domain, function, input_metric, output_measure, privacy_map)


# type: ignore
def then_deintegerize_vec(k: i32) -> Function[Vec[IBig], Vec[TO]]:

    if k == i32.MIN: # |\label{line:check-k}|
        raise ValueError("k must be greater than i32.MIN")
    
    def element_function(x_i):
        return TO.from_rational(x_mul_2k(RBig.from_(x_i), k))
    
    return Function.new(lambda x: [element_function(x_i) for x_i in x])



# type: ignore
def validate_candidates(candidates: list[T]):
    if not candidates: # `\label{empty}`
        raise ValueError("candidates must not be empty")
    
    i1 = iter(candidates)
    i2 = iter(candidates)
    next(i1)

    for c1, c2 in zip(i1, i2): # `\label{windows}`
        cmp = c1.partial_cmp(c2)
        if cmp is None or cmp != Ordering.Less:
            raise ValueError("candidates must be non-null and strictly increasing")



# type: ignore
def make_count(
    input_domain: VectorDomain[AtomDomain[TIA]],
    input_metric: SymmetricDistance
):
    output_domain = AtomDomain(TO) # |\label{line:output-domain}|

    def function(data: Vec[TIA]) -> TO: # |\label{line:TO-output}|
        size = input_domain.size(data) # |\label{line:size}|
        try: # |\label{line:try-catch}|
            return TO.exact_int_cast(size) # |\label{line:exact-int-cast}|
        except FailedCast:
            return TO.MAX_CONSECUTIVE # |\label{line:except-return}|

    output_metric = AbsoluteDistance(TO)

    stability_map = new_stability_map_from_constant(TO.one()) # |\label{line:stability-map}|

    return Transformation(
        input_domain, output_domain, function,
        input_metric, output_metric, stability_map)


# type: ignore
def make_float_to_bigint(
    input_space: tuple[VectorDomain[AtomDomain[T]], LpDistance[P, QI]], k: i32
) -> Transformation[
    VectorDomain[AtomDomain[T]],
    VectorDomain[AtomDomain[IBig]],
    LpDistance[P, QI],
    LpDistance[P, RBig],
]:
    input_domain, input_metric = input_space
    if input_domain.element_domain.nullable():
        raise "input_domain may not contain NaN elements"

    size = input_domain.size
    rounding_distance = get_rounding_distance(k, size, T) # |\label{line:rounding-distance}|

    def elementwise_function(x_i):  # |\label{line:elementwise-function}|
        x_i = RBig.try_from(x_i).unwrap_or(RBig.ZERO)  # |\label{line:try-from}|
        return find_nearest_multiple_of_2k(x_i, k)  # |\label{line:find-nearest}|

    def stability_map(d_in):
        try:
            d_in = RBig.try_from(d_in)
        except Exception:
            raise f"d_in ({d_in}) must be finite"
        return x_mul_2k(d_in + rounding_distance, -k)  # |\label{line:x-mul-2k}|

    return Transformation.new(
        input_domain,
        VectorDomain(  # |\label{line:output-domain}|
            element_domain=AtomDomain.default(IBig),
            size=size,
        ),
        Function.new(lambda x: [elementwise_function(x_i) for x_i in x]),
        input_metric,
        LpDistance.default(),
        StabilityMap.new_fallible(stability_map),
    )



# type: ignore
def make_row_by_row(
    input_domain: DI, 
    input_metric: M, 
    output_row_domain: DO, 
    # a function from input domain's row type to output domain's row type
    row_function: Callable([[DI_RowDomain_Carrier], DO_RowDomain_Carrier])
) -> Transformation:
    
    return make_row_by_row_fallible(
        input_domain, input_metric, output_row_domain, row_function
    )


# type: ignore
def cdp_delta(rho: float, eps: float) -> float:
    """The Rust code may be easier to follow due to more commenting."""
    if rho.is_sign_negative():
        raise ValueError(f"rho ({rho}) must be non-negative")

    if eps.is_sign_negative():
        raise ValueError(f"epsilon ({eps}) must be non-negative")

    if rho.is_zero() or eps.is_infinite():
        return 0.0

    if rho.is_infinite():
        return 1.0

    a_max = eps.inf_add(1.0).inf_div((2.0).neg_inf_mul(rho)).inf_add(2.0)

    a_min = 1.01

    while True:
        diff = a_max - a_min

        a_mid = a_min + diff / 2.0

        if a_mid == a_max or a_mid == a_min:
            break

        # calculate derivative
        deriv = (2.0 * a_mid - 1.0) * rho - eps + a_mid.recip().neg().ln_1p()

        if deriv.is_sign_negative():
            a_min = a_mid
        else:
            a_max = a_mid

    # calculate delta
    a_1 = a_max.inf_sub(1.0)
    ar_e = a_max.inf_mul(rho).inf_sub(eps)

    try:
        t1 = a_1.inf_mul(ar_e)

    except OpenDPException:

        # if t1 is negative, then handle negative overflow by making t1 larger: the most negative finite float
        # making t1 larger makes delta larger, so it's still a valid upper bound
        if a_1.is_sign_negative() != ar_e.is_sign_negative():
            t1 = 1.7976931348623157e308  # f64::MIN
        else:
            raise

    t2 = a_max.inf_mul(a_max.recip().neg().inf_ln_1p())

    delta = t1.inf_add(t2).inf_exp().inf_div((a_max.inf_sub(1.0)))

    # delta is always <= 1
    delta.min(1.0)



# type: ignore
def make_expr_count(
    input_domain: ExprDomain, input_metric: PartitionDistance[MI], expr: Expr
) -> Transformation:
    match expr:  # `\label{match-strategy}`
        case Agg(Count(input, include_nulls)):
            if include_nulls:
                strategy = Strategy.Len
            else:
                strategy = Strategy.Count

        case Function(inputs, function=FunctionExpr.NullCount):
            (input,) = inputs
            strategy = Strategy.NullCount

        case Agg(NUnique(input)):
            strategy = Strategy.NUnique

        case _:
            raise ValueError("expected count, null_count, len, or n_unique expression")

    # check if input is row-by-row
    is_row_by_row = input.make_stable(input_domain.as_row_by_row(), input_metric).is_ok()  # `\label{is-row-by-row}`

    # construct prior transformation
    t_prior = input.make_stable(input_domain, input_metric)  # `\label{t-prior}`
    middle_domain, middle_metric = t_prior.output_space()

    by, margin = middle_domain.context.grouping("count")  # `\label{alignment}`

    output_domain = ExprDomain.new(  # `\label{output-domain}`
        column=SeriesDomain.new(  # `\label{output-series}`
            middle_domain.column.name,
            AtomDomain.default(u32),
        ),
        context=Context.Grouping(
            by=by,
            margin=Margin(
                max_partition_length=1,
                max_num_partitions=margin.max_num_partitions,
                max_partition_contributions=None,
                max_influenced_partitions=margin.max_influenced_partitions,
                public_info=margin.public_info,
            ),
        )
    )

    match strategy:  # `\label{will-count-all}`
        case Strategy.Len:
            will_count_all = is_row_by_row
        case Strategy.Count:
            will_count_all = is_row_by_row and not middle_domain.column.nullable
        case _:
            will_count_all = False

    public_info = margin.public_info if will_count_all else None  # `\label{public-info}`
    
    def function(e: Expr) -> Expr:  # `\label{function}`
        match strategy:
            case Strategy.Count:
                return e.count()
            case Strategy.NullCount:
                return e.null_count()
            case Strategy.Len:
                return e.len()
            case Strategy.NUnique:
                return e.n_unique()

    return t_prior >> Transformation.new(  # `\label{count-trans}`
        middle_domain,
        output_domain,  # `\label{output-domain-usage}`
        Function.then_expr(function),
        middle_metric,
        LpDistance.default(),
        counting_query_stability_map(public_info),  # `\label{stability-map}`
    )



# type: ignore
def counting_query_stability_map(
    public_info: Literal["Keys"] | Literal["Lengths"] | None,
) -> StabilityMap[PartitionDistance[M], LpDistance[P, f64]]:

    if public_info == "Lengths":  # `\label{public-info}`
        return StabilityMap.new(lambda _: 0.)
        
    def norm_map(d_in: f64) -> f64:  # `\label{norm-map}`
        if P == 1:
            return d_in
        if P == 2:
            return d_in.inf_sqrt()
        raise ValueError("unsupported Lp norm. Must be an L1 or L2 norm.")
    
    def stability_map(d_in: tuple[u32, u32, u32]) -> f64:
        l0, l1, l_inf = d_in  # `\label{l01i}`
        l0_p = norm_map(f64.from_(l0))  # `\label{l0-p}`
        l1_p = f64.from_(l1)
        l_inf_p = f64.from_(l_inf)
        return l1_p.total_min(l0_p.inf_mul(l_inf_p))  # `\label{final-bound-impl}`

    return StabilityMap.new_fallible(stability_map)  # `\label{stability-map}`



# Auto-generated. Do not edit!
'''
The ``domains`` module provides functions for creating and using domains.
For more context, see :ref:`domains in the User Guide <domains-user-guide>`.

For convenience, all the functions of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401

__all__ = [
    "_atom_domain_get_bounds_closed",
    "_atom_domain_nan",
    "_domain_equal",
    "_domain_free",
    "_extrinsic_domain_descriptor",
    "_lazyframe_domain_get_columns",
    "_lazyframe_domain_get_margin",
    "_lazyframe_domain_get_series_domain",
    "_lazyframe_from_domain",
    "_member",
    "_option_domain_get_element_domain",
    "_series_domain_get_element_domain",
    "_series_domain_get_name",
    "_series_domain_get_nullable",
    "_vector_domain_get_element_domain",
    "_vector_domain_get_size",
    "array_domain",
    "atom_domain",
    "bitvector_domain",
    "categorical_domain",
    "datetime_domain",
    "domain_carrier_type",
    "domain_debug",
    "domain_type",
    "enum_domain",
    "lazyframe_domain",
    "map_domain",
    "option_domain",
    "series_domain",
    "user_domain",
    "vector_domain",
    "wild_expr_domain",
    "with_margin"
]


def _atom_domain_get_bounds_closed(
    domain: Domain
):
    r"""Retrieve bounds from an AtomDomain<T>

    [_atom_domain_get_bounds_closed in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn._atom_domain_get_bounds_closed.html)

    .. end-markdown

    :param domain: 
    :type domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_domain = py_to_c(domain, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains___atom_domain_get_bounds_closed
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_atom_domain_get_bounds_closed',
            '__module__': 'domains',
            '__kwargs__': {
                'domain': domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _atom_domain_nan(
    domain: Domain
):
    r"""Retrieve whether members of AtomDomain<T> may be NaN.

    [_atom_domain_nan in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn._atom_domain_nan.html)

    .. end-markdown

    :param domain: 
    :type domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_domain = py_to_c(domain, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains___atom_domain_nan
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_atom_domain_nan',
            '__module__': 'domains',
            '__kwargs__': {
                'domain': domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _domain_equal(
    left: Domain,
    right: Domain
) -> bool:
    r"""Check whether two domains are equal.

    .. end-markdown

    :param left: Domain to compare.
    :type left: Domain
    :param right: Domain to compare.
    :type right: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_left = py_to_c(left, c_type=Domain, type_name="AnyDomain")
    c_right = py_to_c(right, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains___domain_equal
    lib_function.argtypes = [Domain, Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_left, c_right), BoolPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_domain_equal',
            '__module__': 'domains',
            '__kwargs__': {
                'left': left, 'right': right
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _domain_free(
    this
):
    r"""Internal function. Free the memory associated with `this`.

    .. end-markdown

    :param this: 
    :type this: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_domains___domain_free
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': '_domain_free',
            '__module__': 'domains',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _extrinsic_domain_descriptor(
    domain: Domain
):
    r"""Retrieve the descriptor value stored in an extrinsic domain.

    .. end-markdown

    :param domain: The ExtrinsicDomain to extract the descriptor from
    :type domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_domain = py_to_c(domain, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains___extrinsic_domain_descriptor
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_domain), ExtrinsicObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_extrinsic_domain_descriptor',
            '__module__': 'domains',
            '__kwargs__': {
                'domain': domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _lazyframe_domain_get_columns(
    lazyframe_domain: Domain
):
    r"""Retrieve the column names of the LazyFrameDomain.

    .. end-markdown

    :param lazyframe_domain: Domain to retrieve the column names from
    :type lazyframe_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_lazyframe_domain = py_to_c(lazyframe_domain, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains___lazyframe_domain_get_columns
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_lazyframe_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_lazyframe_domain_get_columns',
            '__module__': 'domains',
            '__kwargs__': {
                'lazyframe_domain': lazyframe_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _lazyframe_domain_get_margin(
    lazyframe_domain: Domain,
    by
):
    r"""Retrieve the series domain at index 'column`.

    .. end-markdown

    :param lazyframe_domain: Domain to retrieve the SeriesDomain from
    :type lazyframe_domain: Domain
    :param by: grouping columns
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_lazyframe_domain = py_to_c(lazyframe_domain, c_type=Domain, type_name="AnyDomain")
    c_by = py_to_c(by, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["Expr"]))

    # Call library function.
    lib_function = lib.opendp_domains___lazyframe_domain_get_margin
    lib_function.argtypes = [Domain, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_lazyframe_domain, c_by), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_lazyframe_domain_get_margin',
            '__module__': 'domains',
            '__kwargs__': {
                'lazyframe_domain': lazyframe_domain, 'by': by
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _lazyframe_domain_get_series_domain(
    lazyframe_domain: Domain,
    name: str
) -> Domain:
    r"""Retrieve the series domain at index `column`.

    .. end-markdown

    :param lazyframe_domain: Domain to retrieve the SeriesDomain from
    :type lazyframe_domain: Domain
    :param name: Name of the SeriesDomain to retrieve
    :type name: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_lazyframe_domain = py_to_c(lazyframe_domain, c_type=Domain, type_name="AnyDomain")
    c_name = py_to_c(name, c_type=ctypes.c_char_p, type_name="c_char")

    # Call library function.
    lib_function = lib.opendp_domains___lazyframe_domain_get_series_domain
    lib_function.argtypes = [Domain, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_lazyframe_domain, c_name), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': '_lazyframe_domain_get_series_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'lazyframe_domain': lazyframe_domain, 'name': name
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _lazyframe_from_domain(
    domain: Domain
):
    r"""Construct an empty LazyFrame with the same schema as in the LazyFrameDomain.

    This is useful for creating a dummy lazyframe used to write a query plan.

    [_lazyframe_from_domain in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn._lazyframe_from_domain.html)

    .. end-markdown

    :param domain: A LazyFrameDomain.
    :type domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_domain = py_to_c(domain, c_type=Domain, type_name="LazyFrameDomain")

    # Call library function.
    lib_function = lib.opendp_domains___lazyframe_from_domain
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_lazyframe_from_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'domain': domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _member(
    this: Domain,
    val
) -> bool:
    r"""Check membership in a `domain`.

    .. end-markdown

    :param this: The domain to check membership in.
    :type this: Domain
    :param val: A potential element of the domain.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Domain, type_name="AnyDomain")
    c_val = py_to_c(val, c_type=AnyObjectPtr, type_name=domain_carrier_type(this))

    # Call library function.
    lib_function = lib.opendp_domains___member
    lib_function.argtypes = [Domain, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this, c_val), BoolPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_member',
            '__module__': 'domains',
            '__kwargs__': {
                'this': this, 'val': val
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _option_domain_get_element_domain(
    option_domain: Domain
) -> Domain:
    r"""Retrieve the element domain of the option domain.

    .. end-markdown

    :param option_domain: The option domain from which to retrieve the element domain
    :type option_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_option_domain = py_to_c(option_domain, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains___option_domain_get_element_domain
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_option_domain), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': '_option_domain_get_element_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'option_domain': option_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _series_domain_get_element_domain(
    series_domain: Domain
) -> Domain:
    r"""Retrieve the element domain of the series domain.

    [_series_domain_get_element_domain in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn._series_domain_get_element_domain.html)

    .. end-markdown

    :param series_domain: The series domain from which to retrieve the element domain
    :type series_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_series_domain = py_to_c(series_domain, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains___series_domain_get_element_domain
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_series_domain), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': '_series_domain_get_element_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'series_domain': series_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _series_domain_get_name(
    series_domain: Domain
):
    r"""

    .. end-markdown

    :param series_domain: The series domain from which to retrieve the name of elements
    :type series_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_series_domain = py_to_c(series_domain, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains___series_domain_get_name
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_series_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_series_domain_get_name',
            '__module__': 'domains',
            '__kwargs__': {
                'series_domain': series_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _series_domain_get_nullable(
    series_domain: Domain
):
    r"""Retrieve whether elements in members of the domain may be null.

    .. end-markdown

    :param series_domain: The series domain from which to check nullability.
    :type series_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_series_domain = py_to_c(series_domain, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains___series_domain_get_nullable
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_series_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_series_domain_get_nullable',
            '__module__': 'domains',
            '__kwargs__': {
                'series_domain': series_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _vector_domain_get_element_domain(
    vector_domain: Domain
) -> Domain:
    r"""Retrieve the element domain of the vector domain.

    .. end-markdown

    :param vector_domain: The vector domain from which to retrieve the element domain
    :type vector_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_vector_domain = py_to_c(vector_domain, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains___vector_domain_get_element_domain
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_vector_domain), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': '_vector_domain_get_element_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'vector_domain': vector_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _vector_domain_get_size(
    vector_domain: Domain
):
    r"""Retrieve the size of vectors in the vector domain.

    .. end-markdown

    :param vector_domain: The vector domain from which to retrieve the size
    :type vector_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_vector_domain = py_to_c(vector_domain, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains___vector_domain_get_size
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_vector_domain), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_vector_domain_get_size',
            '__module__': 'domains',
            '__kwargs__': {
                'vector_domain': vector_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def array_domain(
    element_domain: Domain,
    width: int
) -> Domain:
    r"""Construct an instance of `ArrayDomain`.
    Can be used as an argument to a Polars series domain.

    .. end-markdown

    :param element_domain: The domain of each element in the array.
    :type element_domain: Domain
    :param width: The width of the array.
    :type width: int
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_element_domain = py_to_c(element_domain, c_type=Domain, type_name="AnyDomain")
    c_width = py_to_c(width, c_type=ctypes.c_uint32, type_name="u32")

    # Call library function.
    lib_function = lib.opendp_domains__array_domain
    lib_function.argtypes = [Domain, ctypes.c_uint32]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_element_domain, c_width), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'array_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'element_domain': element_domain, 'width': width
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def atom_domain(
    bounds = None,
    nan = None,
    T: Optional[RuntimeTypeDescriptor] = None
) -> AtomDomain:
    r"""Construct an instance of `AtomDomain`.

    The domain defaults to unbounded if `bounds` is `None`,
    If `T` is float, `nan` defaults to `true`.

    [atom_domain in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn.atom_domain.html)

    .. end-markdown

    :param bounds: Optional bounds of elements in the domain, if the data type is numeric.
    :param nan: Whether the domain may contain NaN, if the data type is float.
    :param T: The type of the atom.
    :type T: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library

    :example:

    >>> dp.atom_domain(T=float, nan=False)
    AtomDomain(T=f64)

    """
    # Standardize type arguments.
    T = RuntimeType.parse_or_infer(type_name=T, public_example=get_first(bounds))

    # Convert arguments to c types.
    c_bounds = py_to_c(bounds, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=[RuntimeType(origin='Tuple', args=[T, T])]))
    c_nan = py_to_c(nan, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["bool"]))
    c_T = py_to_c(T, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_domains__atom_domain
    lib_function.argtypes = [AnyObjectPtr, AnyObjectPtr, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_bounds, c_nan, c_T), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'atom_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'bounds': bounds, 'nan': nan, 'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def bitvector_domain(
    max_weight = None
) -> Domain:
    r"""Construct an instance of `BitVectorDomain`.

    .. end-markdown

    :param max_weight: The maximum number of positive bits.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_max_weight = py_to_c(max_weight, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["u32"]))

    # Call library function.
    lib_function = lib.opendp_domains__bitvector_domain
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_max_weight), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'bitvector_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'max_weight': max_weight
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def categorical_domain(
    categories = None
) -> Domain:
    r"""Construct an instance of `CategoricalDomain`.
    Can be used as an argument to a Polars series domain.

    .. end-markdown

    :param categories: Optional ordered set of valid string categories
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_categories = py_to_c(categories, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=[RuntimeType(origin='Vec', args=["String"])]))

    # Call library function.
    lib_function = lib.opendp_domains__categorical_domain
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_categories), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'categorical_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'categories': categories
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def datetime_domain(
    time_unit: str = "us",
    time_zone: Optional[str] = None
) -> Domain:
    r"""Construct an instance of `DatetimeDomain`.

    Documentation on valid time zones can be found [in the Polars documentation](https://docs.pola.rs/user-guide/transformations/time-series/timezones/).

    [datetime_domain in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn.datetime_domain.html)

    .. end-markdown

    :param time_unit: One of ``ns``, ``us`` or ``ms``, corresponding to nano-, micro-, and milliseconds
    :type time_unit: str
    :param time_zone: Optional time zone.
    :type time_zone: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_time_unit = py_to_c(time_unit, c_type=ctypes.c_char_p, type_name="str")
    c_time_zone = py_to_c(time_zone, c_type=ctypes.c_char_p, type_name=RuntimeType(origin='Option', args=["str"]))

    # Call library function.
    lib_function = lib.opendp_domains__datetime_domain
    lib_function.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_time_unit, c_time_zone), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'datetime_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'time_unit': time_unit, 'time_zone': time_zone
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def domain_carrier_type(
    this: Domain
) -> str:
    r"""Get the carrier type of a `domain`.

    .. end-markdown

    :param this: The domain to retrieve the carrier type from.
    :type this: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains__domain_carrier_type
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'domain_carrier_type',
            '__module__': 'domains',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def domain_debug(
    this: Domain
) -> str:
    r"""Debug a `domain`.

    .. end-markdown

    :param this: The domain to debug (stringify).
    :type this: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains__domain_debug
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'domain_debug',
            '__module__': 'domains',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def domain_type(
    this: Domain
) -> str:
    r"""Get the type of a `domain`.

    .. end-markdown

    :param this: The domain to retrieve the type from.
    :type this: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains__domain_type
    lib_function.argtypes = [Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'domain_type',
            '__module__': 'domains',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def enum_domain(
    categories
) -> Domain:
    r"""Construct an instance of `EnumDomain`.
    Can be used as an argument to a Polars series domain.

    .. end-markdown

    :param categories: Optional ordered set of string categories
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_categories = py_to_c(categories, c_type=AnyObjectPtr, type_name="Series")

    # Call library function.
    lib_function = lib.opendp_domains__enum_domain
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_categories), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'enum_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'categories': categories
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def lazyframe_domain(
    series_domains
) -> LazyFrameDomain:
    r"""Construct an instance of `LazyFrameDomain`.

    .. end-markdown

    :param series_domains: Domain of each series in the lazyframe.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_series_domains = py_to_c(series_domains, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["SeriesDomain"]))

    # Call library function.
    lib_function = lib.opendp_domains__lazyframe_domain
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_series_domains), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'lazyframe_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'series_domains': series_domains
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def map_domain(
    key_domain: Domain,
    value_domain: Domain
) -> Domain:
    r"""Construct an instance of `MapDomain`.

    .. end-markdown

    :param key_domain: domain of keys in the hashmap
    :type key_domain: Domain
    :param value_domain: domain of values in the hashmap
    :type value_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_key_domain = py_to_c(key_domain, c_type=Domain, type_name="AnyDomain")
    c_value_domain = py_to_c(value_domain, c_type=Domain, type_name="AnyDomain")

    # Call library function.
    lib_function = lib.opendp_domains__map_domain
    lib_function.argtypes = [Domain, Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_key_domain, c_value_domain), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'map_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'key_domain': key_domain, 'value_domain': value_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def option_domain(
    element_domain: Domain,
    D: Optional[RuntimeTypeDescriptor] = None
) -> OptionDomain:
    r"""Construct an instance of `OptionDomain`.

    [option_domain in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn.option_domain.html)

    .. end-markdown

    :param element_domain: 
    :type element_domain: Domain
    :param D: The type of the inner domain.
    :type D: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    D = RuntimeType.parse_or_infer(type_name=D, public_example=element_domain)

    # Convert arguments to c types.
    c_element_domain = py_to_c(element_domain, c_type=Domain, type_name=D)
    c_D = py_to_c(D, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_domains__option_domain
    lib_function.argtypes = [Domain, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_element_domain, c_D), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'option_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'element_domain': element_domain, 'D': D
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def series_domain(
    name: str,
    element_domain: Domain
) -> SeriesDomain:
    r"""Construct an instance of `SeriesDomain`.

    [series_domain in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/domains/fn.series_domain.html)

    .. end-markdown

    :param name: The name of the series.
    :type name: str
    :param element_domain: The domain of elements in the series.
    :type element_domain: Domain
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_name = py_to_c(name, c_type=ctypes.c_char_p, type_name="str")
    c_element_domain = py_to_c(element_domain, c_type=Domain, type_name=None)

    # Call library function.
    lib_function = lib.opendp_domains__series_domain
    lib_function.argtypes = [ctypes.c_char_p, Domain]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_name, c_element_domain), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'series_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'name': name, 'element_domain': element_domain
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def user_domain(
    identifier: str,
    member,
    descriptor = None
) -> Domain:
    r"""Construct a new UserDomain.
    Any two instances of an UserDomain are equal if their string descriptors are equal.
    Contains a function used to check if any value is a member of the domain.


    Required features: `honest-but-curious`

    **Why honest-but-curious?:**

    The identifier must uniquely identify this domain.
    If the identifier is not uniquely identifying,
    then two different domains with the same identifier will chain,
    which can violate transformation stability.

    In addition, the member function must:

    1. be a pure function
    2. be sound (only return true if its input is a member of the domain).

    .. end-markdown

    :param identifier: A string description of the data domain.
    :type identifier: str
    :param member: A function used to test if a value is a member of the data domain.
    :param descriptor: Additional constraints on the domain.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_identifier = py_to_c(identifier, c_type=ctypes.c_char_p, type_name=None)
    c_member = py_to_c(member, c_type=CallbackFnPtr, type_name="bool")
    c_descriptor = py_to_c(descriptor, c_type=ExtrinsicObjectPtr, type_name="ExtrinsicObject")

    # Call library function.
    lib_function = lib.opendp_domains__user_domain
    lib_function.argtypes = [ctypes.c_char_p, CallbackFnPtr, ExtrinsicObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_identifier, c_member, c_descriptor), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'user_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'identifier': identifier, 'member': member, 'descriptor': descriptor
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def vector_domain(
    atom_domain: Domain,
    size = None
) -> VectorDomain:
    r"""Construct an instance of `VectorDomain`.

    .. end-markdown

    :param atom_domain: The inner domain.
    :type atom_domain: Domain
    :param size: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_atom_domain = py_to_c(atom_domain, c_type=Domain, type_name=None)
    c_size = py_to_c(size, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["i32"]))

    # Call library function.
    lib_function = lib.opendp_domains__vector_domain
    lib_function.argtypes = [Domain, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_atom_domain, c_size), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'vector_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'atom_domain': atom_domain, 'size': size
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def wild_expr_domain(
    columns,
    margin = None
) -> Domain:
    r"""Construct a WildExprDomain.


    Required features: `contrib`

    .. end-markdown

    :param columns: descriptors for each column in the data
    :param margin: descriptors for grouped data
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("contrib")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_columns = py_to_c(columns, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["SeriesDomain"]))
    c_margin = py_to_c(margin, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Option', args=["Margin"]))

    # Call library function.
    lib_function = lib.opendp_domains__wild_expr_domain
    lib_function.argtypes = [AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_columns, c_margin), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'wild_expr_domain',
            '__module__': 'domains',
            '__kwargs__': {
                'columns': columns, 'margin': margin
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def with_margin(
    frame_domain: Domain,
    margin
) -> LazyFrameDomain:
    r"""

    .. end-markdown

    :param frame_domain: 
    :type frame_domain: Domain
    :param margin: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_frame_domain = py_to_c(frame_domain, c_type=Domain, type_name=None)
    c_margin = py_to_c(margin, c_type=AnyObjectPtr, type_name="Margin")

    # Call library function.
    lib_function = lib.opendp_domains__with_margin
    lib_function.argtypes = [Domain, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_frame_domain, c_margin), Domain))
    try:
        output.__opendp_dict__ = {
            '__function__': 'with_margin',
            '__module__': 'domains',
            '__kwargs__': {
                'frame_domain': frame_domain, 'margin': margin
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output



# type: ignore
def make_expr_strptime(
    input_domain: WildExprDomain, input_metric: M, expr: Expr
) -> Transformation:
    match expr:  # `\label{match-strategy}`
        case Expr.Function(
            inputs,
            function=FunctionExpr.StringExpr(
                StringFunction.Strptime(to_type, strptime_options)
            ),
        ):
            pass
        case _:
            raise ValueError("expected str.strptime expression")
    input, ambiguous = inputs

    t_prior = input.make_stable(input_domain, input_metric)  # `\label{t-prior}`
    middle_domain, middle_metric = t_prior.output_space()

    if strptime_options.format is None:  # `\label{format}`
        raise ValueError("format must be specified")

    if to_type == DataType.Time and not strptime_options.exact:
        raise ValueError("non-exact not implemented for Time data type")

    # never raise on error
    strptime_options.strict = False  # `\label{non-strict}`

    try:  # `\label{ambiguous}`
        ambig_value = literal_value_of(ambiguous)
    except Exception:
        ambig_value = None
    ambiguous = lit(ambig_value if ambig_value in {"earliest" "latest"} else "null")

    output_domain = middle_domain.clone()
    series_domain = output_domain.column

    # check input and output types
    if series_domain.dtype() != DataType.String:  # `\label{string-check}`
        raise ValueError("str.strptime input dtype must be String")
    
    if to_type not in {DataType.Time, DataType.Datetime, DataType.Date}:  # `\label{to-type-check}`
        raise ValueError("str.strptime output dtype must be Time, Datetime or Date")

    # in Rust, this assigns to the series domain in output_domain
    series_domain.set_dtype(to_type)  # `\label{out-series}`
    series_domain.nullable = True

    def function(expr: Expr) -> Expr:
        return expr.str.strptime(to_type, strptime_options, ambiguous)

    return t_prior >> Transformation.new(  # `\label{strp-trans}`
        middle_domain,
        output_domain,
        Function.then_expr(function),
        middle_metric,
        middle_metric,
        StabilityMap.new(lambda d_in: d_in),
    )



# type: ignore
def score_candidates_constants(size: Optional[u64], alpha: f64) -> tuple[u64, u64, u64]:
    if alpha < 0.0 or 1.0 < alpha:
        return ValueError("alpha must be within [0, 1]")

    alpha_num_exact, alpha_den_exact = RBig.try_from(alpha).into_parts()

    if size is not None:
        # choose the finest granularity that won't overflow
        # must have that size * denom < MAX, so let denom = MAX // size
        alpha_den_approx = u64.MAX.neg_inf_div(size)
    else:
        # default to an alpha granularity of .00001
        u64.exact_int_cast(10_000)

    if alpha_den_exact < UBig.from_(alpha_den_approx):
        alpha_num = u64.try_from(alpha_num_exact.into_parts()[1])
        alpha_den = u64.try_from(alpha_den_exact)
    else:
        # numer = alpha * denom
        alpha_num_approx = u64.round_cast(alpha * f64.round_cast(alpha_den_approx))
        alpha_num, alpha_den = alpha_num_approx, alpha_den_approx

    if size is not None:
        size_limit = size
    else:
        size_limit = u64.MAX.neg_inf_div(alpha_den)

    assert alpha_num <= alpha_den  # `\label{prop-1}`
    size_limit.alerting_mul(alpha_den)  # `\label{prop-2}`

    return alpha_num, alpha_den, size_limit  # `\label{return}`



import opendp.prelude as dp
import pytest



def test_sequential_composition():
    max_influence = 1
    sc_meas = dp.c.make_adaptive_composition(
        input_domain=dp.vector_domain(dp.atom_domain(T=int)),
        input_metric=dp.symmetric_distance(),
        output_measure=dp.max_divergence(),
        d_in=max_influence,
        d_mids=[0.2, 0.3, 0.4, 0.7],
    )

    sc_qbl: dp.Queryable = sc_meas([1] * 200)

    print("sc_qbl", sc_qbl)
    sum_query = (
        sc_meas.input_space
        >> dp.t.then_clamp((0, 10))
        >> dp.t.then_sum()
        >> dp.m.then_laplace(100.0)
    )

    print("sc_qbl(sum_query)", sc_qbl(sum_query))

    exact_sum = (
        sc_meas.input_space
        >> dp.t.then_clamp((0, 10))
        >> dp.t.then_sum()
    )
    print("exact sum:", exact_sum)
    exact_sum_sc_qbl = sc_qbl(
        exact_sum
        >> dp.c.make_adaptive_composition(
            input_domain=exact_sum.output_domain,
            input_metric=exact_sum.output_metric,
            output_measure=dp.max_divergence(),
            d_in=exact_sum.map(max_influence),
            d_mids=[0.2, 0.09],
        )
    )
    noise_query = exact_sum.output_space >> dp.m.then_laplace(200.0)

    print("child release:", exact_sum_sc_qbl(noise_query))
    print("child release:", exact_sum_sc_qbl(noise_query))
    print("root release: ", sc_qbl(sum_query))


def test_sequential_composition_approxdp():
    max_influence = 1
    sc_meas = dp.c.make_adaptive_composition(
        input_domain=dp.vector_domain(dp.atom_domain(T=int)),
        input_metric=dp.symmetric_distance(),
        output_measure=dp.fixed_smoothed_max_divergence(),
        d_in=max_influence,
        d_mids=[(1.0, 1e-6), (1.0, 1e-6)],
    )

    sc_qbl: dp.Queryable = sc_meas([1] * 200)

    input_space = dp.vector_domain(dp.atom_domain(T=int)), dp.symmetric_distance()

    sum_meas = (
        input_space
        >> dp.t.then_clamp((0, 10))
        >> dp.t.then_sum()
        >> dp.m.then_gaussian(100.0)
    )
    sum_meas = dp.c.make_fix_delta(dp.c.make_zCDP_to_approxDP(sum_meas), 1e-6)
    sc_qbl(sum_meas)


def test_plugin_queryable_int():
    def transition(query):
        assert query == 2
        return query + 1
    qbl = dp.new_queryable(transition, int, int)
    assert qbl(2) == 3


def test_plugin_queryable_list():
    def transition(query, _is_internal):
        assert query == [2, 3]
        return query[-1]
    qbl = dp.new_queryable(transition, "Vec<i32>", int)
    assert qbl([2, 3]) == 3


def test_plugin_queryable_error():
    def transition(_query, _is_internal):
        raise ValueError("test clean stack trace")
    qbl = dp.new_queryable(transition, "Vec<i32>", int)

    with pytest.raises(dp.OpenDPException):
        qbl([2, 3])

    with pytest.raises(TypeError):
        qbl(2)


def test_fully_adaptive_composition():
    max_influence = 1
    space = dp.vector_domain(dp.atom_domain(T=int)), dp.symmetric_distance()
    o_comp = space >> dp.c.then_fully_adaptive_composition(dp.max_divergence())
    assert space == o_comp.input_space

    assert str(o_comp) == """Odometer(
    input_domain   = VectorDomain(AtomDomain(T=i32)),
    input_metric   = SymmetricDistance(),
    output_measure = MaxDivergence)"""

    qbl_comp: dp.OdometerQueryable = o_comp([1] * 200)
    assert qbl_comp.privacy_loss(max_influence) == 0.0

    assert str(qbl_comp) == "OdometerQueryable(Q=AnyMeasurement, QB=u32)"
    m_sum = space >> dp.t.then_clamp((0, 10)) >> dp.t.then_sum() >> dp.m.then_laplace(100.)

    # evaluating
    assert isinstance(qbl_comp(m_sum), int)
    assert qbl_comp.privacy_loss(max_influence) == m_sum.map(max_influence)

    m_lap = dp.m.make_laplace(dp.atom_domain(T=int), dp.absolute_distance(T=int), 200.)
    t_sum = space >> dp.t.then_clamp((0, 10)) >> dp.t.then_sum()
    m_sum_compositor = t_sum >> dp.c.then_adaptive_composition(
        output_measure=dp.max_divergence(),
        d_in=t_sum.map(max_influence),
        d_mids=[0.2, 0.09]
    )
    qbl_summed = qbl_comp.invoke(m_sum_compositor)
    # it's slightly larger, checking greater than will do
    assert qbl_comp.privacy_loss(max_influence) > m_sum.map(max_influence) + 0.2 + 0.09

    assert isinstance(qbl_summed(m_lap), int) # child release
    assert isinstance(qbl_summed(m_lap), int) # child release
    assert isinstance(qbl_comp(m_sum), int) # root release

    # it's slightly larger, checking greater than will do
    assert qbl_comp.privacy_loss(1) > m_sum.map(max_influence) * 2 + 0.2 + 0.09


def test_odometer_supporting_elements():
    o_ac = dp.c.make_fully_adaptive_composition(
        input_domain=dp.vector_domain(dp.atom_domain(T=int)),
        input_metric=dp.symmetric_distance(),
        output_measure=dp.max_divergence(),
    )

    assert isinstance(o_ac.invoke([]), dp.OdometerQueryable)
    assert o_ac.input_domain == dp.vector_domain(dp.atom_domain(T=int))
    assert o_ac.input_metric == dp.symmetric_distance()
    assert o_ac.output_measure == dp.max_divergence()
    assert o_ac.input_space == (dp.vector_domain(dp.atom_domain(T=int)), dp.symmetric_distance())
    assert o_ac.input_distance_type == dp.u32
    assert o_ac.output_distance_type == dp.f64
    assert o_ac.input_carrier_type == dp.Vec[dp.i32]


def test_privacy_filter():
    m_filter = dp.c.make_privacy_filter(
        dp.c.make_fully_adaptive_composition(
            input_domain=dp.vector_domain(dp.atom_domain(T=int)),
            input_metric=dp.symmetric_distance(),
            output_measure=dp.max_divergence(),
        ),
        d_in=1,
        d_out=2.0
    )

    qbl_filter = m_filter([1, 2, 3, 4, 5])

    m_count = dp.t.make_count(*m_filter.input_space) >> dp.m.then_laplace(1.0)

    qbl_filter(m_count)
    assert qbl_filter.privacy_loss(1) == 1.0
    qbl_filter(m_count)
    assert qbl_filter.privacy_loss(1) == 2.0




from opendp.extras.polars import Bound
from opendp.mod import ExtrinsicDistance, FrameDistance, SymmetricIdDistance
import opendp.prelude as dp
import pytest


def test_l01inf_distance():
    domain = dp.vector_domain(dp.atom_domain(T=float))
    metric = dp.l01inf_distance(dp.symmetric_distance())
    assert metric != str(metric)
    trans = dp.t.make_user_transformation(
        domain,
        metric,
        domain,
        metric,
        function=lambda x: x,
        stability_map=lambda d_in: d_in,
    )

    assert trans.map((3, 4, 3)) == (3, 4, 3)


def test_identifier_distance():
    pl = pytest.importorskip("polars")

    metric = dp.unit_of(contributions=1, identifier="A")[0]
    assert metric.identifier.meta.serialize() == pl.col.A.meta.serialize()

    group_bound = Bound(by=[], per_group=1)
    metric = dp.unit_of(contributions=[group_bound], identifier="C")[0]
    assert isinstance(metric, FrameDistance)
    assert isinstance(metric.inner_metric, SymmetricIdDistance)
    assert metric.inner_metric.identifier.meta.serialize() == pl.col.C.meta.serialize()


def test_frame_distance():
    pytest.importorskip("polars")
    metric = dp.frame_distance(dp.symmetric_distance())
    # not a valid domain/metric pairing, but behind honest-but-curious
    t_dummy = dp.t.make_identity(dp.atom_domain(T=int), metric)
    d_in = [Bound(by=["A"], per_group=10)]
    assert t_dummy.map(d_in) == d_in


def test_group_bound():
    pl = pytest.importorskip("polars")
    left = Bound(by=[pl.col.A, "B"], per_group=10)
    right = Bound(by=[pl.col.B, "A"], per_group=10)
    assert left == right
    assert not left == Bound(by=[pl.col.B], per_group=10)
    assert not left == str(right)


def test_user_metric_total_cmp_native_distance():
    m_comp = dp.c.make_adaptive_composition(
        input_domain=dp.atom_domain(T=bool),
        input_metric=dp.user_distance("user distance"),
        output_measure=dp.max_divergence(),
        d_in=1,
        d_mids=[1.0],
    )
    assert m_comp.map(0) == 1.0
    assert m_comp.map(1) == 1.0
    with pytest.raises(
        dp.OpenDPException,
        match="d_in from the privacy map must be no greater than the d_in",
    ):
        m_comp.map(2)
    with pytest.raises(dp.OpenDPException, match="not comparable"):
        m_comp.map(float("nan"))

def test_user_metric_total_cmp_custom_distance():
    class Dist:
        def __lt__(self, other):
            raise ValueError("comparison failed!")

    m_comp = dp.c.make_adaptive_composition(
        input_domain=dp.atom_domain(T=bool),
        input_metric=dp.user_distance("user distance", ["other", "data"]),
        output_measure=dp.max_divergence(),
        d_in=Dist(),
        d_mids=[1.0],
    )
    with pytest.raises(dp.OpenDPException, match="comparison failed!"):
        m_comp.map(Dist())

    assert isinstance(m_comp.input_metric, ExtrinsicDistance)
    assert m_comp.input_metric.cast(list) == ["other", "data"]
    assert m_comp.input_metric.descriptor == ["other", "data"]

    with pytest.raises(ValueError, match="metric descriptor must be a int"):
        m_comp.input_metric.cast(int)


def test_bound():
    pytest.importorskip("polars")
    zero_way = dp._get_bound([Bound(by=["A"], per_group=2)], ["A", "B"])
    assert zero_way == Bound(by=["A", "B"], per_group=2)
